@@ -4,19 +4,35 @@
 Authors:
     * Carl Simon Adorf <simon.adorf@epfl.ch>
 """
+import asyncio
 import getpass
 import logging
 import socket
 from dataclasses import dataclass, field
 from pathlib import Path
+from shutil import rmtree
 from textwrap import wrap
 
 import click
 import docker
+from packaging.version import parse
 from tabulate import tabulate
 
-from .core import APPLICATION_ID, LOGGER, AiidaLabInstance, Config, Profile
-from .util import get_docker_client, spinner, webbrowser_available
+from .core import (
+    APPLICATION_ID,
+    DEFAULT_PORT,
+    LOGGER,
+    AiidaLabInstance,
+    Config,
+    Profile,
+)
+from .util import (
+    confirm_with_value,
+    get_docker_client,
+    get_latest_version,
+    spinner,
+    webbrowser_available,
+)
 from .version import __version__
 
 MSG_MOUNT_POINT_CONFLICT = """Warning: There is at least one other running
@@ -113,6 +129,15 @@ def cli(verbose):
 
     LOGGER.info(f"Configuration file path: {_application_config_path()}")
 
+    latest_version = get_latest_version(timeout=0.1)
+    if latest_version and latest_version > parse(__version__):
+        click.secho(
+            f"A new version of aiidalab-launch is available: {latest_version} (installed: {__version__})",
+            fg="yellow",
+        )
+        if "pipx" in __file__:
+            click.secho("Run `pipx upgrade aiidalab-launch` to update.", fg="yellow")
+
 
 @cli.command()
 def version():
@@ -157,9 +182,23 @@ def show_profile(app_state, profile):
 
 @profiles.command("add")
 @click.argument("profile")
+@click.option(
+    "--port",
+    type=click.IntRange(min=1, max=65535),
+    help=(
+        "Specify port on which this instance will be exposed. The default port "
+        "is chosen such that it does not conflict with any currently configured "
+        "profiles."
+    ),
+)
+@click.option(
+    "--home-mount",
+    type=click.Path(file_okay=False),
+    help="Specify the path to be mounted as home directory.",
+)
 @pass_app_state
 @click.pass_context
-def add_profile(ctx, app_state, profile):
+def add_profile(ctx, app_state, port, home_mount, profile):
     """Add a new AiiDAlab profile to the configuration."""
     try:
         app_state.config.get_profile(profile)
@@ -168,7 +207,21 @@ def add_profile(ctx, app_state, profile):
     else:
         raise click.ClickException(f"Profile with name '{profile}' already exists.")
 
-    new_profile = Profile(name=profile)
+    # Determine next available port or use the one provided by the user.
+    configured_ports = [prof.port for prof in app_state.config.profiles if prof.port]
+    port = port or (max(configured_ports, default=-1) + 1) or DEFAULT_PORT
+    # Determine home mount path unless provided.
+    home_mount = home_mount or str(Path.home() / f"aiidalab-{profile}")
+
+    try:
+        new_profile = Profile(
+            name=profile,
+            port=port,
+            home_mount=home_mount,
+        )
+    except ValueError as error:  # invalid profile name
+        raise click.ClickException(error)
+
     app_state.config.profiles.append(new_profile)
     app_state.config.save(_application_config_path())
     click.echo(f"Added profile '{profile}'.")
@@ -179,14 +232,29 @@ def add_profile(ctx, app_state, profile):
 @profiles.command("remove")
 @click.argument("profile")
 @click.option("--yes", is_flag=True, help="Do not ask for confirmation.")
+@click.option("-f", "--force", is_flag=True, help="Proceed, ignoring any warnings.")
 @pass_app_state
-def remove_profile(app_state, profile, yes):
+def remove_profile(app_state, profile, yes, force):
     """Remove a AiiDAlab profile from the configuration."""
     try:
         profile = app_state.config.get_profile(profile)
     except ValueError:
         raise click.ClickException(f"Profile with name '{profile}' does not exist.")
     else:
+        if not force:
+            instance = AiidaLabInstance(client=app_state.docker_client, profile=profile)
+            status = asyncio.run(instance.status())
+            if status not in (
+                instance.AiidaLabInstanceStatus.DOWN,
+                instance.AiidaLabInstanceStatus.CREATED,
+                instance.AiidaLabInstanceStatus.EXITED,
+            ):
+                raise click.ClickException(
+                    f"The instance associated with profile '{profile.name}' "
+                    "is still running. Use the -f/--force option to remove the "
+                    "profile anyways."
+                )
+
         if yes or click.confirm(
             f"Are you sure you want to remove profile '{profile.name}'?"
         ):
@@ -227,7 +295,7 @@ def set_default_profile(app_state, profile):
         click.echo(f"Set default profile to '{profile}'.")
 
 
-def _find_mount_point_conflict(client, profile, other_profiles):
+async def _find_mount_point_conflict(client, profile, other_profiles):
     """Find running instances with the same home mount point.
 
     To protect users from inadvertently starting a second profile with the same
@@ -239,10 +307,167 @@ def _find_mount_point_conflict(client, profile, other_profiles):
             other_profile != profile
             and Path(other_profile.home_mount).resolve()
             == Path(profile.home_mount).resolve()
-            and AiidaLabInstance(client=client, profile=other_profile).status()
-            is not AiidaLabInstance.AiidaLabInstanceStatus.DOWN
         ):
-            yield other_profile
+            status = await AiidaLabInstance(
+                client=client, profile=other_profile
+            ).status()
+            if status is not AiidaLabInstance.AiidaLabInstanceStatus.DOWN:
+                yield other_profile
+
+
+async def _async_start(
+    app_state, profile, restart, wait, pull, no_browser, show_ssh_help, force
+):
+    # Check for potential mount point conflicts.
+    instance = AiidaLabInstance(client=app_state.docker_client, profile=profile)
+
+    if not force:
+        conflict = False
+        with spinner("Check for potential conflicts...", delay=0.1):
+            async for p in _find_mount_point_conflict(
+                app_state.docker_client,
+                profile,
+                app_state.config.profiles,
+            ):
+                conflict = True
+                break
+
+        if conflict:
+            msg_warn = MSG_MOUNT_POINT_CONFLICT.format(home_mount=profile.home_mount)
+            click.confirm(
+                click.style("\n".join(wrap(msg_warn)), fg="yellow"),
+                abort=True,
+            )
+
+    # Obtain image (either via pull or local).
+    if pull:
+        try:
+            msg = (
+                f"Downloading image '{instance.profile.image}', this may take a while..."
+                if instance.image is None
+                else f"Downloading latest version of '{instance.profile.image}'..."
+            )
+            with spinner(msg):
+                instance.pull()
+        except RuntimeError as error:
+            raise click.ClickException(str(error))
+    elif instance.image is None:
+        raise click.ClickException(
+            f"Unable to find image '{profile.image}'. "
+            "Try to use '--pull' to pull the image prior to start."
+        )
+
+    # Check if image has changed.
+    assert instance.image is not None
+    image_changed = (
+        instance.container and instance.container.image.id != instance.image.id
+    )
+    mounts_changed = (
+        instance.container
+        and f"{instance.profile.container_name()}_conda"
+        not in (mount.get("Name") for mount in instance.container.attrs["Mounts"])
+    )
+    recreate = image_changed or mounts_changed
+
+    try:
+
+        InstanceStatus = instance.AiidaLabInstanceStatus  # local alias for brevity
+
+        status = await instance.status()
+        if status in (
+            InstanceStatus.DOWN,
+            InstanceStatus.CREATED,
+            InstanceStatus.EXITED,
+        ):
+            if recreate:
+                with spinner("Recreating container..."):
+                    instance.remove()
+                    instance.create()
+            with spinner("Starting container..."):
+                instance.start()
+        elif status is InstanceStatus.UP and restart:
+            with spinner("Restarting container..."):
+                if any(instance.configuration_changes()):
+                    instance.stop()
+                    instance.remove()
+                    instance.start()
+                else:
+                    instance.restart()
+        elif status is InstanceStatus.UP and not restart:
+            if any(instance.configuration_changes()):
+                click.secho(
+                    "Container is already running, however the configuration "
+                    "has changed. A restart with --restart is recommended.",
+                    fg="yellow",
+                )
+            else:
+                click.echo(
+                    "Container was already running, use --restart to restart it.",
+                    err=True,
+                )
+        elif status is InstanceStatus.STARTING:
+            click.echo("Container is already starting up...", err=True)
+
+    except docker.errors.APIError as error:
+        LOGGER.debug(f"Error during startup: {error}")
+        if instance.profile.port and "port is already allocated" in str(error):
+            raise click.ClickException(
+                f"Port {instance.profile.port} is already allocated, choose another port "
+                f"for example, by editing the profile: aiidalab-launch profiles edit {instance.profile.name}"
+            )
+        raise click.ClickException("Startup failed due to an unexpected error.")
+    except asyncio.TimeoutError:
+        raise click.ClickException(
+            "AiiDAlab instance did not start up within the excepted wait period."
+        )
+    except Exception as error:
+        raise click.ClickException(f"Unknown error occurred: {error}")
+    else:
+        if wait:
+            try:
+                with spinner("Waiting for AiiDAlab instance to get ready..."):
+                    echo_logs = asyncio.create_task(instance.echo_logs())
+                    await asyncio.wait_for(instance.wait_for_services(), timeout=wait)
+                    echo_logs.cancel()
+            except asyncio.TimeoutError:
+                raise click.ClickException(
+                    f"AiiDAlab instance did not start up within the provided wait period ({wait})."
+                )
+            except RuntimeError:
+                raise click.ClickException(
+                    "The AiiDAlab instance failed to start. Consider to inspect "
+                    "the container output logs by increasing the output "
+                    "verbosity with 'aiidalab-launch -vvv start'."
+                )
+            msg_startup = (
+                MSG_STARTUP_SSH
+                if (show_ssh_help or not webbrowser_available())
+                else MSG_STARTUP
+            )
+            url = instance.url()
+            click.secho(
+                msg_startup.format(
+                    url=url,
+                    home_mount=instance.profile.home_mount,
+                    system_user=instance.profile.system_user,
+                    user=getpass.getuser(),
+                    port=instance.profile.port,
+                    hostname=socket.getfqdn(),
+                ).lstrip(),
+                fg="green",
+            )
+            if not no_browser and webbrowser_available():
+                if click.confirm(
+                    "Do you want to open AiiDAlab in the browser now?", default=True
+                ):
+                    click.launch(url)
+
+        else:
+            click.secho(
+                "Use 'aiidalab-launch status' to check the AiiDAlab instance "
+                "status and URL to open it.",
+                fg="green",
+            )
 
 
 @cli.command()
@@ -285,145 +510,9 @@ def _find_mount_point_conflict(client, profile, other_profiles):
 )
 @pass_app_state
 @with_profile
-def start(app_state, profile, restart, wait, pull, no_browser, show_ssh_help, force):
+def start(*args, **kwargs):
     """Start an AiiDAlab instance on this host."""
-
-    instance = AiidaLabInstance(client=app_state.docker_client, profile=profile)
-
-    # Check for potential mount point conflicts.
-    if not force:
-        with spinner("Check for potential conflicts...", delay=0.1):
-            conflict = any(
-                _find_mount_point_conflict(
-                    app_state.docker_client,
-                    profile,
-                    app_state.config.profiles,
-                )
-            )
-        if conflict:
-            msg_warn = MSG_MOUNT_POINT_CONFLICT.format(home_mount=profile.home_mount)
-            click.confirm(
-                click.style("\n".join(wrap(msg_warn)), fg="yellow"),
-                abort=True,
-            )
-
-    # Obtain image (either via pull or local).
-    if pull:
-        try:
-            msg = (
-                f"Downloading image '{instance.profile.image}', this may take a while..."
-                if instance.image is None
-                else f"Downloading latest version of '{instance.profile.image}'..."
-            )
-            with spinner(msg):
-                instance.pull()
-        except RuntimeError as error:
-            raise click.ClickException(str(error))
-    elif instance.image is None:
-        raise click.ClickException(
-            f"Unable to find image '{profile.image}'. "
-            "Try to use '--pull' to pull the image prior to start."
-        )
-
-    # Check if image has changed.
-    assert instance.image is not None
-    recreate = instance.container and instance.container.image.id != instance.image.id
-
-    try:
-
-        InstanceStatus = instance.AiidaLabInstanceStatus  # local alias for brevity
-
-        status = instance.status()
-        if status in (
-            InstanceStatus.DOWN,
-            InstanceStatus.CREATED,
-            InstanceStatus.EXITED,
-        ):
-            if recreate:
-                with spinner("Recreating container..."):
-                    instance.remove()
-                    instance.create()
-            with spinner("Starting container..."):
-                instance.start()
-        elif status is InstanceStatus.UP and restart:
-            with spinner("Restarting container..."):
-                if recreate:
-                    instance.stop()
-                    instance.remove()
-                instance.start()
-        elif status is InstanceStatus.UP and not restart:
-            if recreate:
-                click.secho(
-                    "Container is already running, however the image has changed. "
-                    "A restart with --restart is recommended.",
-                    fg="yellow",
-                )
-            else:
-                click.echo(
-                    "Container was already running, use --restart to restart it.",
-                    err=True,
-                )
-        elif status is InstanceStatus.STARTING:
-            click.echo("Container is already starting up...", err=True)
-
-    except TimeoutError:
-        raise click.ClickException(
-            f"AiiDAlab instance did not start up within the provided wait period ({wait})."
-        )
-    except docker.errors.APIError as error:
-        LOGGER.debug(f"Error during startup: {error}")
-        if instance.profile.port and "port is already allocated" in str(error):
-            raise click.ClickException(
-                f"Port {instance.profile.port} is already allocated, choose another port "
-                f"for example, by editing the profile: aiidalab-launch profiles edit {instance.profile.name}"
-            )
-        raise click.ClickException("Startup failed due to an unexpected error.")
-    except TimeoutError:
-        raise click.ClickException(
-            "AiiDAlab instance did not start up within the excepted wait period."
-        )
-    except Exception as error:
-        raise click.ClickException(f"Unknown error occurred: {error}")
-    else:
-        if wait:
-            try:
-                with spinner("Waiting for AiiDAlab instance to get ready..."):
-                    instance.wait_for_services(timeout=wait)
-            except RuntimeError:
-                raise click.ClickException(
-                    "The AiiDAlab instance failed to start. Consider to inspect "
-                    "the container output logs by increasing the output "
-                    "verbosity with 'aiidalab-launch -v start'."
-                )
-            url = instance.url()
-            msg_startup = (
-                MSG_STARTUP_SSH
-                if (show_ssh_help or not webbrowser_available())
-                else MSG_STARTUP
-            )
-            click.secho(
-                msg_startup.format(
-                    url=instance.url(),
-                    home_mount=instance.profile.home_mount,
-                    system_user=instance.profile.system_user,
-                    user=getpass.getuser(),
-                    port=instance.profile.port,
-                    hostname=socket.getfqdn(),
-                ).lstrip(),
-                fg="green",
-            )
-            if not no_browser and webbrowser_available():
-                if click.confirm(
-                    "Do you want to open AiiDAlab in the browser now?", default=True
-                ):
-                    click.launch(url)
-
-        else:
-            click.secho(
-                "Use 'aiidalab-launch status' to check the AiiDAlab instance "
-                "status and URL to open it.",
-                fg="green",
-            )
+    asyncio.run(_async_start(*args, **kwargs))
 
 
 @cli.command()
@@ -445,8 +534,12 @@ def start(app_state, profile, restart, wait, pull, no_browser, show_ssh_help, fo
 def stop(app_state, profile, remove, timeout):
     """Stop an AiiDAlab instance on this host."""
     instance = AiidaLabInstance(client=app_state.docker_client, profile=profile)
-    status = instance.status()
-    if status is instance.AiidaLabInstanceStatus.UP:
+    status = asyncio.run(instance.status())
+    if status not in (
+        instance.AiidaLabInstanceStatus.DOWN,
+        instance.AiidaLabInstanceStatus.CREATED,
+        instance.AiidaLabInstanceStatus.EXITED,
+    ):
         with spinner("Stopping AiiDAlab...", final="stopped."):
             instance.stop(timeout=timeout)
     if remove:
@@ -479,14 +572,17 @@ def status(app_state):
     # Collect status for each profile
     headers = ["Profile", "Container", "Status", "Mount", "URL"]
     rows = []
-    instances = (
+    instances = [
         AiidaLabInstance(client=client, profile=profile)
         for profile in app_state.config.profiles
-    )
+    ]
+
+    async def fetch_status(instances):
+        results = await asyncio.gather(*(instance.status() for instance in instances))
+        return list(zip(instances, results))
 
     with spinner("Collecting status info...", delay=0.5):
-        for instance in instances:
-            instance_status = instance.status()
+        for instance, instance_status in asyncio.run(fetch_status(instances)):
             rows.append(
                 [
                     instance.profile.name,
@@ -531,10 +627,10 @@ def exec(ctx, profile, cmd, privileged, forward_exit_code, wait):
     try:
         if wait:
             with spinner("Waiting for AiiDAlab to get ready...", delay=0.5):
-                instance.wait_for_services(timeout=60)
+                asyncio.run(asyncio.wait_for(instance.wait_for_services(), timeout=60))
         with spinner("Send command to container...", delay=1.0):
             exec_id = instance.exec_create(" ".join(cmd), privileged=privileged)
-    except RuntimeError:
+    except (RuntimeError, asyncio.TimeoutError):
         raise click.ClickException("AiiDAlab instance is not available. Is it running?")
 
     output = app_state.docker_client.api.exec_start(exec_id, stream=True)
@@ -549,5 +645,62 @@ def exec(ctx, profile, cmd, privileged, forward_exit_code, wait):
             ctx.fail(f"Command failed with exit code: {result['ExitCode']}")
 
 
+@cli.command()
+@click.option(
+    "--apps",
+    is_flag=True,
+    help="Only remove installed apps. The default apps will be installed on the next instance start.",
+)
+@click.option("--yes", is_flag=True, help="Do not ask for confirmation.")
+@with_profile
+@pass_app_state
+def reset(app_state, profile, apps, yes):
+    """Reset an AiiDAlab instance."""
+    # Check (and abort) in case that the instance is running.
+    instance = AiidaLabInstance(client=app_state.docker_client, profile=profile)
+    status = asyncio.run(instance.status())
+    if status not in (
+        instance.AiidaLabInstanceStatus.DOWN,
+        instance.AiidaLabInstanceStatus.CREATED,
+        instance.AiidaLabInstanceStatus.EXITED,
+    ):
+        raise click.ClickException(
+            f"The instance associated with profile '{profile.name}' "
+            "is still running. Please stop it prior to reset."
+        )
+
+    click.secho(
+        f"Resetting instance for profile '{profile.name}'. This action cannot be undone!",
+        err=True,
+        fg="red",
+    )
+
+    if not yes:
+        confirm_with_value(
+            profile.name, "Please enter the name of the profile to continue", abort=True
+        )
+
+    def rmtree_(path: Path) -> None:
+        if path.exists():
+            try:
+                rmtree(path)
+            except Exception as error:
+                raise click.ClickException(
+                    f"Encountered error while trying to remove '{path}': {error}"
+                )
+
+    if apps:
+        click.echo(
+            "Removing apps directory. Default apps will be installed on next start."
+        )
+        rmtree_(Path(profile.home_dir) / "apps")
+    else:
+        click.echo("Removing container and associated volumes.")
+        instance.remove()
+
+        click.echo(f"Removing home directory '{profile.home_mount}'.")
+        rmtree_(Path(profile.home_mount))
+
+
 if __name__ == "__main__":
     cli()
